open Lab79BsSmartPracticeModels;
open Js.Promise;
open Lab79BsKnex2;

[@bs.val] external batchSize_: Js.Nullable.t(int) = "process.env.BATCH_SIZE";

let batchSize =
  switch (Js.Nullable.toOption(batchSize_)) {
  | None => 50
  | Some(x) => x
  };

exception UnexpectedValidateResult;

let callAttemptor = (channel, task, taskAttemptId) =>
  switch (channel) {
  | `OutboundCall =>
    Model.knex |> Lab79BsSpCall.Make.attempt(~task, ~taskAttemptId)
  | `SMSChatbot => Lab79BsSpChatbot.Attempt.make(task, taskAttemptId)
  };

let tryAttempt = (task: Tasks.Schema.t) => {
  let maybeExpectedChannel =
    task |> Tasks.Schema.expectedNextAttemptChannelGet;
  switch (
    Tasks.Schema.(
      task |> idGet,
      task |> targetTypeGet,
      task |> taskTypeGet,
      task |> targetIdGet,
    )
  ) {
  | (Some(taskId), Some(targetType), Some(taskType), Some(targetId)) =>
    switch (maybeExpectedChannel) {
    | None =>
      let errInfo = "No expected Channel specified";
      Model.Task.Update.resetState(~taskId, ~toState=`Abnormal, ~info=errInfo)
      |> then_(_ => resolve(Some(errInfo)));
    | Some(channel) =>
      Validate.validate(task)
      |> then_(validateResult =>
           switch (validateResult) {
           | None =>
             TaskAttempts.Insert.fromTask(
               Model.knex,
               ~taskId,
               ~targetType,
               ~targetId,
               ~channel,
             )
             |> then_(attemptId =>
                  Model.Task.Update.resetState(
                    ~taskId,
                    ~toState=`Attempting,
                    ~info="attempting",
                  )
                  |> then_(_ =>
                       callAttemptor(channel, task, attemptId)
                     )
                )

           | Some(failure) =>
             let failureInfo = failure |> Validate.ValidateFailures.infoGet;
             Model.Task.Update.resetState(
               ~taskId,
               ~toState=failure |> Validate.ValidateFailures.taskStateToGet,
               ~info=failureInfo,
             )
             |> then_(_ => resolve(Some(failureInfo)));
           }
         )
    }
  | _ => raise(Exceptions.MissingAttr)
  };
};

/**
  * Running per channel per task type
  */
let batchCheckTasks = channelId => {
  let gotChannelToTargetCounts =
    Model.TaskAttempt.CountByTarget.fromChannel(channelId);
  let gotChannelInfo = TaskChannels.Get.byId(Model.knex, ~id=channelId);
  let gotPendingTasks =
    Model.Task.Get.queryDispatchedByChannelId(batchSize, channelId);

  all3((gotChannelToTargetCounts, gotChannelInfo, gotPendingTasks))
  |> then_(((channelCounts, channelInfo, pendingTasks)) =>
       switch (
         TaskChannels.Schema.(
           channelInfo |> maxConcurrentTaskAttemptsAcrossTargetsGet,
           channelInfo |> maxConcurrentTaskAttemptsPerTargetGet,
         )
       ) {
       | (Some(limitationAcrossTargets), Some(limitationPerTarget)) =>
         let totalLeft =
           limitationAcrossTargets
           - Array.fold_left(
               (x, y) => x + y,
               0,
               Js.Dict.values(channelCounts),
             );

         let pendingTaskCount = Array.length(pendingTasks);

         (
           totalLeft <= 0 ?
             [] :
             Array.sub(
               pendingTasks,
               0,
               totalLeft < pendingTaskCount ? totalLeft : pendingTaskCount,
             )
             |> Array.fold_left(
                  (result, task) =>
                    switch (task |> Tasks.Schema.targetIdGet) {
                    | None => raise(Exceptions.MissingAttr)
                    | Some(targetId) =>
                      let count_ = Js.Dict.get(channelCounts, targetId);
                      let count =
                        switch (count_) {
                        | None => 0
                        | Some(count) => count
                        };

                      count < limitationPerTarget ?
                        {
                          Js.Dict.set(channelCounts, targetId, count + 1);
                          [task, ...result];
                        } :
                        result;
                    },
                  [],
                )
         )
         |> resolve;
       | _ => reject(Exceptions.MissingAttr)
       }
     );
};

let run = taskChannelId =>
  batchCheckTasks(taskChannelId)
  |> then_(tasks =>
       tasks |> List.map(tryAttempt) |> resolve
     )
  |> then_(_ => Knex2.destroy(Model.knex) |> resolve)
  |> catch(err => {
    Js.log(err);
    Knex2.destroy(Model.knex);
    () |> resolve;
  });

Js.log("Task Attempter Start.");
run("9e40deac-a834-4023-ad32-61d42b7799bd") |> ignore;
Js.log("Task Attempter End.");