open Lab79BsSmartPracticeModels;
exception NoSuchTaskType;

module SP_Tasks = Lab79SmartPracticeTasks;

module ValidateFailures = {
  [@bs.deriving abstract]
  type t = {
    taskStateTo: Enums.taskState,
    info: string,
  };
};

let isPhysicianAvailable = (task, configParams) => {
  let dateStart =
    MomentRe.Moment.add(
      Util.RelTime.biasedNow(0.0 /*** TODO: pysician's timezone */),
      ~duration=
        configParams
        |> TaskTypes.Configurations.Params.startGet
        |> Util.RelTime.toMomentDuration,
    );
  let dateEnd =
    MomentRe.Moment.add(
      Util.RelTime.biasedNow(0.0 /*** TODO: pysician's timezone */),
      ~duration=
        configParams
        |> TaskTypes.Configurations.Params.end_Get
        |> Util.RelTime.toMomentDuration,
    );
  let limitation =
    configParams |> TaskTypes.Configurations.Params.limitationGet;

  switch (task |> Tasks.Schema.paramsGet) {
  | None =>
    Some(
      ValidateFailures.t(~taskStateTo=`Abnormal, ~info="Missing task params"),
    )
    |> Js.Promise.resolve

  | Some(params_) =>
    let params =
      params_ |> SP_Tasks.ScheduleAppointmentTask.Params.fromJsonDict;
    switch (params |> SP_Tasks.ScheduleAppointmentTask.Params.ecwGet) {
    | None =>
      Some(
        ValidateFailures.t(~taskStateTo=`Abnormal, ~info="missing ECW Info"),
      )
      |> Js.Promise.resolve
    | Some(ecw) =>
      EcwAvailability.Physician.Query.check(
        ~accountId=
          switch (task |> Tasks.Schema.accountIdGet) {
          | None => raise(Exceptions.MissingAttr)
          | Some(accountId) => accountId
          },
        ~startDate=dateStart,
        ~endDate=dateEnd,
        ~duration=
          params |> SP_Tasks.ScheduleAppointmentTask.Params.durationMinutesGet,
        ~ecw,
      )
      |> Js.Promise.then_(result => {
           let past = result |> EcwAvailability.Physician.Query.Result.pastGet;
           let futureAvailable =
             result
             |> EcwAvailability.Physician.Query.Result.futureGet
             |> EcwAvailability.Physician.Query.Result.Future.availableGet;
           let futureUnavailable =
             result
             |> EcwAvailability.Physician.Query.Result.futureGet
             |> EcwAvailability.Physician.Query.Result.Future.unavailableGet;
           let futureTotal = futureAvailable + futureUnavailable;
           let bookedTotal = past + futureUnavailable;
           let allInTotal = past + futureTotal;
           let isAvailable =
             switch (
               configParams |> TaskTypes.Configurations.Params.checkTypeGet
             ) {
             | `PercentageTotal =>
               float(bookedTotal) /. float(allInTotal) <= limitation
             | `PercentageLeft =>
               float(futureUnavailable) /. float(futureTotal) <= limitation
             | `AmountTotal => float(futureAvailable) >= limitation
             };
           (
             isAvailable ?
               None :
               Some(
                 ValidateFailures.t(
                   ~taskStateTo=`NeedReschedule,
                   ~info=
                     "Need reschedule since physician not available at given time window",
                 ),
               )
           )
           |> Js.Promise.resolve;
         })
    };
  };
};

let mayNotMatchSpecialContext = (taskType: Seed.taskTypes) =>
  switch (taskType) {
  | `AppointmentReminder => ((task, config) => Js.Promise.resolve(None))
  | `ScheduleAppointment => (
      (task, config) =>
        isPhysicianAvailable(
          task,
          config |> TaskTypes.Configurations.Schema.paramsGet,
        )
    )
  };

let validate = task =>
  switch (task |> Tasks.Schema.expectedLatestAttemptGet) {
  | None => None |> Js.Promise.resolve
  | Some(latest) =>
    MomentRe.momentWithDate(latest)
    |> MomentRe.Moment.isBefore(
         Util.RelTime.biasedNow(0.0 /*** TODO: pysician's timezone */),
       ) ?
      Some(
        ValidateFailures.t(
          ~taskStateTo=`NeedReschedule,
          ~info="Task Outdated",
        ),
      )
      |> Js.Promise.resolve :
      TaskTypes.Configurations.Get.fromTask(Model.knex, ~task)
      |> Js.Promise.then_(config =>
           switch (task |> Tasks.Schema.taskTypeGet) {
           | None =>
             Js.Promise.resolve(
               Some(
                 ValidateFailures.t(
                   ~taskStateTo=`Abnormal,
                   ~info="Missing task type",
                 ),
               ),
             )
           | Some(taskType) =>
             mayNotMatchSpecialContext(taskType, task, config)
           }
         )
  };